
local math = math
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local SDL = require 'SDL'
local gl = require 'gl'; local GL = gl
local Fonts = require 'Fonts'
local Node = require 'Node'
local Textures = require 'Textures'
local Event = require 'Event'
local Cursor = require 'Cursor'
local Music = require 'yarivan.Music'   -- TODO move dependency elsewhere
local FS = FS
local _G = _G

module(...)

local Menu = _M
Node.Root:subclass(Menu)

Menu.alpha = 0
Menu.fade = 0.2

function Menu:initialize()
  self.focus = self.focus or Cursor.AnimationState:new{ animFn =
    function (self, percent) return percent end }
  self.focus:go({ self.focusTo or 1 })
  self.focusTo = nil
end

function Menu:displayContent()
  local w2, h2 = _G.WIDTH/2, _G.HEIGHT/2
  if self:getLoop().parentloop then self:getLoop().parentloop:display() end
  if self.background then
    local tex = self.background
    gl.Color(1, 1, 1, self.alpha)
    gl.BindTexture(GL.TEXTURE_2D, tex.name)
    gl.Begin(GL.QUADS)
      gl.TexCoord(0, 0); gl.Vertex(0, 0)
      gl.TexCoord(tex.wratio, 0); gl.Vertex(tex.origw, 0)
      gl.TexCoord(tex.wratio, tex.hratio); gl.Vertex(tex.origw, tex.origh)
      gl.TexCoord(0, tex.hratio); gl.Vertex(0, tex.origh)
    gl.End()
  end
  if self.background2 then
    local tex = self.background2
    gl.Color(1, 1, 1, self.alpha)
    gl.BindTexture(GL.TEXTURE_2D, tex.name)
    gl.Begin(GL.QUADS)
      gl.TexCoord(0, 0); gl.Vertex(0, 0)
      gl.TexCoord(tex.wratio, 0); gl.Vertex(tex.origw, 0)
      gl.TexCoord(tex.wratio, tex.hratio); gl.Vertex(tex.origw, tex.origh)
      gl.TexCoord(0, tex.hratio); gl.Vertex(0, tex.origh)
    gl.End()
  end
  -- TODO reduce code duplication with MapView:displayContent()
  local message = self.items
  local focus = self.focus:getCurrent()[1]
  local font = Fonts['fonts/ANTIQUEN.ttf:16']
  local bh = font.height * #message
  local bw = 0
  for i,v in ipairs(message) do bw = math.max(bw, font:prepare(v)) end
  bh = bh + 20; bw = bw + 20
  local fy = h2+font.height*(#message/2-focus)-(font.height-font.ascent)
  gl.BindTexture(GL.TEXTURE_2D, 0)
  gl.Color(0, 0, 0, 0.5*self.alpha)
  gl.Begin(GL.QUADS)
    gl.Vertex(w2-bw/2-10, h2-bh/2); gl.Vertex(w2-bw/2-10, h2+bh/2)
    gl.Vertex(w2+bw/2+10, h2+bh/2); gl.Vertex(w2+bw/2+10, h2-bh/2)
    gl.Color(0.7, 0, 0, 0.5*self.alpha)
    gl.Vertex(w2-bw/2, fy); gl.Vertex(w2-bw/2, fy+font.height)
    gl.Vertex(w2+bw/2, fy+font.height); gl.Vertex(w2+bw/2, fy)
  gl.End()
  gl.Color(1, 1, 1, self.alpha)
  for i,v in ipairs(message) do
    gl.PushMatrix()
    gl.Translate(math.floor(w2-font:prepare(v)/2),
      math.floor(h2+font.height*(#message/2-i)), 0)
    font:print(v)
    gl.PopMatrix()
  end
end

local musiclist

function Menu:update()
  self.focus:update()
  self.alpha = math.max(0, math.min(1, self.alpha + self.fade))
  if self.fade ~= 0 and self.alpha == 0 then self.fade = 0; self:finish() end
  if not musiclist then musiclist = {}; for k,v in pairs(FS.listFiles('music/')) do musiclist[#musiclist+1] = 'music/'..k end end
  Music.update(musiclist)
end

function Menu:finish()
  self:getLoop().running = false
end

function Menu:eventKey(ev)
  local now = self.focus.dest[1]
  if ev.down and ev.sym == SDL.K_DOWN then
    if now == #self.items then self.focus:go({ 1 }) else self.focus:go({ now+1 }, 4) end end
  if ev.down and ev.sym == SDL.K_UP then
    if now == 1 then self.focus:go({ #self.items }) else self.focus:go({ now-1 }, 4) end end
  if ev.down and (ev.sym == SDL.K_RETURN or ev.sym == SDL.K_KP_ENTER) then
    self.hooks[self.focus.dest[1]](self, self.focus.dest[1]) end
  if ev.down and ev.sym == SDL.K_ESCAPE then
    if self.escape then
      self.hooks[self.escape](self, self.escape)
    else
      self.fade = -0.2
    end
  end
end

