
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local math = math
local FS = FS
local SDL = SDL
local _G = _G

module(...)

local cache = {}

function play(fname, ...)
  -- _G._SOUND is whether the sound engine could be initialized.
  -- _G.SOUND is a user config variable telling if sound is enabled.
  if not (_G._SOUND and _G.SOUND) then return end
  local rw = FS.loadRW(fname)
  cache[fname] = cache[fname] or SDL.LoadMUS_RW(rw); rw:disown()
  cache[fname]:PlayMusic(...)
end

function stop()
  SDL.HaltMusic()
end

local current = nil
local currentN = nil

function update(list)
  if not (_G._SOUND and _G.SOUND) then return end
  if list ~= current then
    current = list
    currentN = nil
    SDL.HaltMusic()
  end
  if current and #current > 0 and not SDL.PlayingMusic() and not SDL.PausedMusic() then
    local newN = math.random(#current)
    while newN == currentN and #current > 1 do
      newN = math.random(#current) end
    currentN = newN
    play(current[currentN])
  end
end

