
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local math = math
local Textures = require 'Textures'
local gl = require 'gl'; local GL = gl
local Actor = require 'yarivan.Actor'

module(...)

local Projectile = _M
Actor:subclass(Projectile)

Projectile.size = 0.2
Projectile.speed = 11
Projectile.direction = 1

function Projectile:loadTextures(info)
  -- TODO avoid needless code duplication with Actor:loadTextures()
  self.textures = { dlists = {}, delay = info.delay, frames = info.frames }
  for frame = 1, info.frames do
    local tex = Textures.load(info.base..frame..'.png{^2}')
    local dlist = gl.GenLists(1)
    gl.NewList(dlist, GL.COMPILE)
      gl.Translate(-(info.rel_x or 0), info.rel_y or 0, 0)
      gl.BindTexture(GL.TEXTURE_2D, tex.name)
      gl.Begin(GL.QUADS)
        gl.TexCoord(0, 0); gl.Vertex(0, 0)
        gl.TexCoord(tex.wratio, 0); gl.Vertex(tex.origw, 0)
        gl.TexCoord(tex.wratio, tex.hratio); gl.Vertex(tex.origw, tex.origh)
        gl.TexCoord(0, tex.hratio); gl.Vertex(0, tex.origh)
      gl.End()
    gl.EndList()
    for dir = 0, 8 do
      self.textures[dir..frame] = tex
      self.textures.dlists[dir..frame] = dlist
    end
  end
end

function Projectile:update()
  if self:move(self.direction, self.speed) < 1 then
    for k,v in ipairs(self.parent) do
      local threshold = 0.6
      if v.isEnemy and Actor.squaresIntersect(v.ix, v.iy, v.size,
          self.ix-threshold, self.iy-threshold, self.size+2*threshold) then
        v:damage()
        v:throwBack(self.direction, 11, 4)
        break
      end
    end
    self.parent:removeChild(self)
  end
end

function Projectile:collidesWith(actor)
  return Actor.collidesWith(self, actor) and
    (actor.isEnemy or (self.parent.arena and actor.player))
end

function Projectile:move(direction, distance)
  self.direction = direction
  self.animationTime = (self.animationTime or 0) + 1
  if distance == 0 then self.animationTime = nil; return end
  local dx, dy = self.parent:getDirection(direction, distance)
  local moveDistance = self:findCollision(self.ix, self.iy, dx, dy)
  self:setPosition(self.ix+moveDistance*dx, self.iy+moveDistance*dy)
  return moveDistance
end

function Projectile:canAccessField(ix, iy)
  for i = 1, #self.parent.data.layers do
    local tileinfo = self.parent.cache[i][iy] and self.parent.cache[i][iy][ix]
    if i == 1 and not tileinfo then return false end
    if tileinfo and tileinfo.noaccess and not tileinfo.projectile then return false end
  end
  return true
end

