
local Node = require 'Node'
local Mask = require 'Mask'
local Event = require 'Event'
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice

module(...)

local WindowManager = _M
Node:subclass(WindowManager)

WindowManager.keyfocus = nil

function WindowManager:eventFilter(ev)
  if ev.type == 'Key' and not ev.wmOverride and self.keyfocus then
    ev.target = self.keyfocus; ev.wmOverride = true; Event.send(ev)
    return true
  end
  if (ev.type == 'MouseButton' or ev.type == 'MouseButtonAway') and
    ev.down and ev.target ~= self.keyfocus then self:setKeyFocus(nil) end
end

function WindowManager:setKeyFocus(obj)
  if self.keyfocus == obj then return end
  local oldfocus = self.keyfocus; self.keyfocus = obj
  -- when node:lostKeyFocus() returns true, it overrides the focus change
  if oldfocus and oldfocus:lostKeyFocus() then self.keyfocus = oldfocus
  elseif self.keyfocus then self.ekyfocus:gotKeyFocus() end
end

function WindowManager:displayContent()
  Mask.Stencil.prepareTopLevel()
  Node.displayContent(self)
  Mask.Stencil.cleanupTopLevel()
end

