
require 'MainLoop'
require 'Textures'
require 'Fonts'
require 'TextInput'

local _G = _G
local gl, glu, SDL = gl, glu, SDL
local Fonts, TextInput, MainLoop = Fonts, TextInput, MainLoop
local Class = Class

module(...)

local Console = _M
MainLoop:instantiate(Console)

Console.font = 'fonts/DejaVuSansMono.ttf:10'
Console.texture = nil

Console.command = nil
Console.input = nil
Console.scroll = 0
Console.histpos = 0

function Console:initialize()
  MainLoop.initialize(self)
  self.history = self.history or {}
  self.output = self.output or {}
  self.input = self.input or TextInput:new{ font = Fonts[self.font] }
end

function Console:write(...)
  for l in _G.table.concat({...}):gmatch('[^\n]+') do
    self.output[#self.output+1] = l
  end
end

function Console:display()
  gl.Clear(gl.COLOR_BUFFER_BIT + gl.DEPTH_BUFFER_BIT)
  if self.parentloop then self.parentloop:display() end
  gl.PushAttrib(gl.ALL_ATTRIB_BITS)
  gl.MatrixMode(gl.PROJECTION)
  gl.PushMatrix()
  gl.LoadIdentity()
  gl.Ortho(0, _G.WIDTH, 0, _G.HEIGHT, -1, 1)
  gl.MatrixMode(gl.MODELVIEW)
  gl.PushMatrix()
  -- TODO maybe disable some more things (stencil etc.) -- if needed
  gl.Disable(gl.DEPTH_TEST)
  gl.Disable(gl.LIGHTING)
  gl.Enable(gl.TEXTURE_2D)
  gl.BlendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA); gl.Enable(gl.BLEND)
  gl.LoadIdentity()
  gl.Translate(30, 30, 0)
  gl.Color(0, 0, 0, 0.5); gl.BindTexture(gl.TEXTURE_2D, 0)
  Fonts.DrawBoundingBox(Fonts[self.font],
    Fonts[self.font]:prepare(self.input.text) + 3)
  gl.Color(1, 1, 1, 1)
  gl.PushMatrix(); self.input:display(); gl.PopMatrix()
  for i = 0, 30 do
    local id = #self.output - i - self.scroll
    if self.output[id] then
      gl.Translate(0, Fonts[self.font].height, 0)
      gl.Color(0, 0, 0, 0.5); gl.BindTexture(gl.TEXTURE_2D, 0)
      Fonts.DrawBoundingBox(Fonts[self.font], self.output[id])
      gl.Color(1, 1, 1, 1)
      gl.PushMatrix(); Fonts[self.font]:print(self.output[id]); gl.PopMatrix()
    end
  end
  gl.PopMatrix()
  gl.MatrixMode(gl.PROJECTION)
  gl.PopMatrix()
  gl.MatrixMode(gl.MODELVIEW)
  gl.PopAttrib()
end

function Console:update()
  self.input:update()
  if self.command then
    local c = self.command; self.command = nil
    local f, err = _G.loadstring('return '..c, '=debug')
    if not f then f, err = _G.loadstring(c, '=debug') end
    if not f then self:write(err); return end
    local e = { loop = self.parentloop,
      self = self, cwrite = function (...) return self:write(...) end }
    _G.setmetatable(e, { __index = _G, __newindex = _G }); _G.setfenv(f, e)
    local _, result = _G.pcall(f)   -- TODO multret
    if result ~= nil and _G.type(result) ~= 'string' then
      result = _G.inspect.object(result, _G.inspect.indented) end
    self:write(result)
  end
end

function Console:key(ev)
  if not ev.down then return end
  if ev.sym == SDL.K_ESCAPE or ev.sym == SDL.K_BACKQUOTE then self.running = false
  elseif ev.sym == SDL.K_PAGEUP then
    if self.output[#self.output - self.scroll - 10] then
      self.scroll = self.scroll + 10 end
  elseif ev.sym == SDL.K_PAGEDOWN then
    if self.output[#self.output - self.scroll + 10] then
      self.scroll = self.scroll - 10 end
  elseif ev.sym == SDL.K_UP or ev.sym == SDL.K_DOWN then
    if #self.history > 0 then
      self.histpos = self.histpos + (ev.sym == SDL.K_UP and -1 or 1)
      while self.histpos < 1 do
        self.histpos = self.histpos + #self.history end
      while self.histpos > #self.history do
        self.histpos = self.histpos - #self.history end
      self.input.text = self.history[self.histpos] or ''
      self.input.cursor = self.input.text:ulen()
    end
  elseif ev.sym == SDL.K_RETURN or ev.sym == SDL.K_KP_ENTER then
    self.command = self.input.text
    self.history[#self.history+1] = self.input.text
    self:write(self.input.text)
    self.histpos, self.scroll = 1, 0
    self.input.text, self.input.cursor = '', 0
  else
    self.input:key(ev)
  end
end
