
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local unpack = unpack
local SDL = require 'SDL'
local gl = require 'gl'
local Object = require 'Object'
local Event = require 'Event'
local Node = require 'Node'
local Selection = require 'Selection'
local _G = _G   -- TODO move _G.HEIGHT elsewhere

module(...)

-- TODO move AnimationState to a separate file
AnimationState = Object:subclass()

AnimationState.t = nil
AnimationState.length = nil
AnimationState.old = nil
AnimationState.dest = nil

function AnimationState:animFn(time_percent)
  --return ((1 - _G.math.cos(time_percent * _G.math.pi)) / 2)
  return _G.math.cos((1 - time_percent) * _G.math.pi/2)
end

function AnimationState:getCurrent()
  -- IMPORTANT TODO use frameFraction
  if self.t and self.t >= self.length then self.t, self.old = nil, nil end
  local result, dest = {}, self.dest
  if not self.t or not self.old then return dest end
  local progress = self:animFn(self.t/self.length)   -- (0..1) clampf
  for k,v in pairs(self.old) do result[k] = v + progress*(dest[k] - v) end
  return result
end

function AnimationState:go(dest, length)
  if not length then   -- go to dest immediately
    self.old, self.t = nil, nil
  else
    self.old = self:getCurrent()
    self.t = 1
    self.length = length
  end
  self.dest = dest
end

function AnimationState:update()
  if self.t then self.t = self.t + 1 end
end

local Cursor = _M
Node:subclass(Cursor)

--Cursor.layer = 20   -- TODO use another layer mechanism in scion.CombatMode
Cursor.doesDisplay = false
Cursor.moveTime = 10
Cursor.pos = nil
Cursor.aux = nil

function Cursor:initialize()
  Node.initialize(self)
  self.pos = AnimationState:new{}
  self.aux = AnimationState:new{}
  -- self.pos should always have SDL.GetMouseState() as dest
  function self.pos:__set_dest(d)
    self._dest = d
    SDL.WarpMouse(d[1], _G.HEIGHT - d[2])
  end
  function self.pos:__get_dest()
    -- maybe (TODO really?) SDL.WarpMouse() won't work immediately, so don't
    -- use SDL.GetMouseState() (through cur_x,y) until self.t > 1
    if self.t == 1 then return self.dest end
    return { Event.mouse.cur_x, _G.HEIGHT - Event.mouse.cur_y }
  end
end

function Cursor:hide()
  self.doesDisplay = false
end

function Cursor:moveAux(...)
  self.aux:go({...}, self.doesDisplay and self.moveTime)
  self.doesDisplay = true
end

function Cursor:move(x, y, ...)
  self.pos:go({x, y}, self.doesDisplay and self.moveTime)
  self.aux:go({...}, self.doesDisplay and self.moveTime)
  self.doesDisplay = true
end

function Cursor:update()
  self.pos:update()
  self.aux:update()
end

function Cursor:display()
  gl.PushMatrix()
  Node.display(self)
  gl.PopMatrix()
end

function Cursor:displayContent()
  if Selection.active then return end
  local cur_x, cur_y = unpack(self.pos:getCurrent())
  gl.LoadIdentity()
  gl.Translate(cur_x, cur_y, 0)
  gl.BindTexture(gl.TEXTURE_2D, 0)
  gl.Color(1, 1, 1)
  gl.LineWidth(2)
  gl.Begin(gl.LINES)
    gl.Vertex(-10, 0); gl.Vertex(10, 0)
    gl.Vertex(0, -10); gl.Vertex(0, 10)
  gl.End()
end

