
local _G = _G
require 'ftgl'  -- for rw:font()
local gl = require 'gl'
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local type, tonumber, setmetatable = type, tonumber, setmetatable
local Textures = require 'Textures'
local FS = FS

module(...)

setmetatable(_M, { __index = function (_M, name)
  local kerning = not name:match('&$')
  local size = assert(tonumber(name:match(':(%d+)&?$')),
    "no font size specified")
  local file = name:gsub(':%d+&?$', '')
  -- TODO should we provide variable texture size?
  _M[name] = assert(FS.loadRW(file)):font(size, 256,
    Textures.generateAndBindName(), kerning)
  return _M[name]
end })

function GetDimensions(font)
  -- TODO: FTGL has imprecise bbox information
  return font.ascent, font.ascent - font.height
end

function DrawBoundingBox(font, right)
  local top, bottom = GetDimensions(font)
  if type(right) ~= 'number' then right = font:prepare(right) end
  gl.Begin(gl.QUADS)
    gl.Vertex(0, top); gl.Vertex(0, bottom)
    gl.Vertex(right, bottom); gl.Vertex(right, top)
  gl.End()
end

function GetCursorPosition(font, text, cursor_x)
  -- TODO with binary search, this could be O(log N)
  local len, old_pos, current_pos = text:ulen(), nil, 0
  for i = 1, len do
    old_pos, current_pos = current_pos, font:prepare(text:usub(1, i))
    if cursor_x < (old_pos + current_pos)/2 then return i-1 end
  end
  return len
end
