
local _G = _G
local assert, unpack = assert, unpack
local gl, GL = gl, GL

module(...)

Stencil = {}

function Stencil.overlay(dontPush)
  -- draw a quad on the whole viewport
  -- (used e.g. for globally decreasing stencil buffer values)
  if not dontPush then
    gl.PushAttrib(GL.ALL_ATTRIB_BITS)
    gl.MatrixMode(GL.MODELVIEW); gl.PushMatrix()
  end
  gl.Disable(GL.DEPTH_TEST)
  gl.DepthMask(0)
  gl.LoadIdentity()
  gl.MatrixMode(GL.PROJECTION)
  gl.PushMatrix(); gl.LoadIdentity()
  gl.Begin(GL.QUADS)
    gl.Vertex(-1, -1, -1); gl.Vertex(1, -1, -1)
    gl.Vertex(1, 1, -1); gl.Vertex(-1, 1, -1)
  gl.End()
  gl.PopMatrix()
  if not dontPush then
    gl.MatrixMode(GL.MODELVIEW); gl.PopMatrix()
    gl.PopAttrib()
  end
end

function Stencil.prepareTopLevel()
  gl.Enable(GL.STENCIL_TEST)
  gl.StencilFunc(GL.EQUAL, 0, -1)
end

function Stencil.cleanupTopLevel()
  gl.Disable(GL.STENCIL_TEST)
end

function Stencil.push(shape)
  local depth = gl.GetBoolean(GL.DEPTH_TEST, 1); gl.Disable(GL.DEPTH_TEST)
  gl.StencilOp(GL.KEEP, GL.KEEP, GL.INCR)
  gl.ColorMask(0, 0, 0, 0); shape:displayStencil(); gl.ColorMask(1, 1, 1, 1)
  if depth then gl.Enable(GL.DEPTH_TEST) end
  gl.StencilFunc(GL.EQUAL, gl.GetInteger(GL.STENCIL_REF, 1) + 1, -1)
  gl.StencilOp(GL.KEEP, GL.KEEP, GL.KEEP)
end

function Stencil.pop()
  -- restore old stencil buffer values
  gl.StencilOp(GL.KEEP, GL.KEEP, GL.DECR)
  gl.ColorMask(0, 0, 0, 0); Stencil.overlay(); gl.ColorMask(1, 1, 1, 1)
  gl.StencilFunc(GL.EQUAL, gl.GetInteger(GL.STENCIL_REF, 1) - 1, -1)
end


Scissor = {}   -- TODO test it (and btw, is it needed at all?)

local scissorStack = { n = 0 }

function Scissor.push(x, y, w, h)
  local now = gl.GetInteger(GL.SCISSOR_BOX, 4)
  local x2, y2 = x+w-1, y+h-1
  local nowx2, nowy2 = now[1]+now[3]-1, now[2]+now[4]-1
  if x < now[1] then x = now[1] end; if y < now[2] then y = now[2] end
  if x2 < nowx2 then x2 = nowx2 end; if y2 < nowy2 then y2 = nowy2 end
  scissorStack.n = scissorStack.n + 1; scissorStack[scissorStack.n] = now
  gl.Scissor(x, y, x2-x+1, y2-y+1)
end

function Scissor.pop()
  assert(scissorStack.n > 0, "scissor stack is empty")
  gl.Scissor(unpack(scissorStack[scissorStack.n]))
  scissorStack.n = scissorStack.n - 1
end
