
local _G = _G
local SDL = SDL

module(...)

local musicChannel = nil

local cache = {}

function load(fname)
  if not _G._SOUND then return nil end
  if not cache[fname] then
    local rw = _G.FS.load(fname)
    cache[fname] = SDL.LoadWAV_RW(rw, true); rw:disown()
  end
  return cache[fname]
end

function play(fname, ...)
  -- _G._SOUND is whether the sound engine could be initialized.
  -- _G.SOUND is a user config variable telling if sound is enabled.
  if not (_G._SOUND and _G.SOUND) then return end
  musicChannel = SDL.PlayChannel(nil, load(fname), ...)
end

function stop()
  if musicChannel then SDL.HaltChannel(musicChannel) end
  musicChannel = nil
end

local current = nil
local currentN = nil

function update(list)
  if not (_G._SOUND and _G.SOUND) then return end
  if list ~= current then
    current = list
    currentN = nil
    stop()
  end
  if current and #current > 0 and not (musicChannel and
    (SDL.Playing(musicChannel) or SDL.Paused(musicChannel))) then
    local newN = _G.math.random(#current)
    while newN == currentN and #current > 1 do
      newN = _G.math.random(#current) end
    currentN = newN
    play(current[currentN])
  end
end
