
require 'Textures'

local _G = _G
local math, table = math, table
local ipairs, type = ipairs, type
local gl = gl
local Textures = Textures

module(...)

local Node = _M
_G.Class(Node)

function Node:initialize()
end

function Node:gotFocus()
end

function Node:lostFocus()
  -- when this returns true, it overrides focus change and retains focus
end

function Node:children()
  local result = {}
  for i,v in ipairs(self) do
    if type(v) ~= 'table' then v = self[v] end
    result[i] = v
  end
  return result
end

function Node:display()
  self:displayContent()
end

function Node:displayContent()
  for i,v in ipairs(self:children()) do v:display() end
end

function Node:update()
  for i,v in ipairs(self:children()) do v:update() end
end

function Node:addChild(name, node)
  if node then self[name] = node end
  self[#self+1] = name
end

function Node:removeChild(name)
  self[name] = nil
  for i,v in ipairs(self) do
    if v == name then table.remove(self, i); break end
  end
end

-- TODO getLoop() must die
function Node:getLoop()
  return self.mainloop or (self.parent and self.parent:getLoop())
end

function Node:event(ev)
  local handler = 'event' .. ev.type
  if not self[handler] then return false end
  return self[handler](self, ev)
end

function Node:eventFilter(ev)
  return false
end


-- a helper class that uses eventFilter() to allow access to the console.
Root = Node:instantiate()

function Root:eventFilter(ev)
  if ev.type == 'Key' and ev.down and ev.sym == _G.SDL.K_BACKQUOTE then
    self.mainloop:subLoop(_G.require'Console'); return true end  -- TODO ugly
end


Moving = Node:instantiate()

Moving.floorPosition = false
Moving.isOffset = true   -- if false, gl.LoadIdentity() is used
Moving.x, Moving.y, Moving.z = 0, 0, 0
Moving.dx, Moving.dy, Moving.dz = 0, 0, 0

function Moving:display()
  gl.PushMatrix()
  if not self.isOffset then gl.LoadIdentity() end
  local ff = self:getLoop().frameFraction
  local x, y, z = self.x+self.dx*ff, self.y+self.dy*ff, self.z+self.dz*ff
  if self.floorPosition then
    x, y, z = math.floor(x), math.floor(y), math.floor(z) end
  gl.Translate(x, y, z)
  Node.display(self)
  gl.PopMatrix()
end

function Moving:update()
  Node.update(self)
  if self.x and self.y and self.z then
    self.x = self.x + self.dx
    self.y = self.y + self.dy
    self.z = self.z + self.dz
  end
end
