
local _G = _G
local SDL, gl = SDL, gl

module(...)

local TextInput = _M
_G.Class(TextInput)

TextInput.text = ''
TextInput.cursor = nil
TextInput.blink = 0
TextInput.font = nil
TextInput.selected = false

-- BIG TODO: it would be neat if TextInput could be controlled by mouse

function TextInput:initialize()
  _G.assert(self.font)
  self.font:prepare(self.text)
  if not self.cursor then self.cursor = self.text:ulen() end
  SDL.EnableUNICODE(true)
  SDL.EnableKeyRepeat(SDL.DEFAULT_REPEAT_DELAY, SDL.DEFAULT_REPEAT_INTERVAL)
end

function TextInput:key(ev)
  if not ev.down then return false end
  local unicode = ev.unicode ~= 127 and ev.unicode ~= 0 and ev.unicode
  if self.blink then self.blink = 0 end
  if ev.sym == SDL.K_HOME then
    self.selected = false
    self.cursor = 0
  elseif ev.sym == SDL.K_END then
    self.selected = false
    self.cursor = self.text:ulen()
  elseif ev.sym == SDL.K_LEFT then
    self.selected = false
    if self.cursor <= 0 then return false end
    self.cursor = self.cursor - 1
  elseif ev.sym == SDL.K_RIGHT then
    self.selected = false
    if self.cursor >= self.text:ulen() then return false end
    self.cursor = self.cursor + 1
  elseif (ev.sym == SDL.K_DELETE or ev.sym == SDL.K_BACKSPACE) and self.selected then
    self.text, self.cursor, self.selected = '', 0, false
  elseif ev.sym == SDL.K_BACKSPACE then
    return self:key{ down = true, sym = SDL.K_LEFT } and
      self:key{ down = true, sym = SDL.K_DELETE }
  elseif ev.sym == SDL.K_DELETE then
    if self.cursor >= self.text:ulen() then return false end
    self.text = self.text:usub(1, self.cursor) ..
      self.text:usub(self.cursor + 2, -1)
  elseif unicode and unicode >= 32 then
    if self.selected then self.text, self.cursor = '', 0 end
    self.selected = false
    self.text = self.text:usub(1, self.cursor) ..
      _G.string.uchar(ev.unicode) .. self.text:usub(self.cursor + 1, -1)
    self.cursor = self.cursor + 1
    self.font:prepare(self.text)   -- load new glyphs
  else
    return false
  end
  return true
end

function TextInput:displayCursor()
  if not self.blink or self.blink < 1 then
    gl.PushMatrix()
    gl.Translate(self.font:prepare(self.text:usub(1, self.cursor)), 0, 0)
    gl.LineWidth(1)
    gl.BindTexture(gl.TEXTURE_2D, 0)
    gl.Begin(gl.LINES)
      gl.Vertex(1, self.font.ascent - self.font.height)
      gl.Vertex(1, self.font.ascent)
    gl.End()
    gl.PopMatrix()
  end
end

function TextInput:display()
  self:displayCursor()
  self.font:print(self.text)
end

TextInput.displayContent = TextInput.display

function TextInput:update()
  if self.blink then self.blink = (self.blink + (1/15)) % 2 end
end
