
local _G = _G
local setmetatable, type = setmetatable, type
local sqrt, max = math.sqrt, math.max

module(...)

local Vector = _M
setmetatable(_M, {
  __call = function (f, o) return setmetatable(o or {}, _M) end })

local names = { x = 1, y = 2, z = 3, w = 4 }
function _M.__index(t, k)
  if names[k] then return t[names[k]] end
  if type(k) == 'number' then return 0 end
  return _M[k] end

function _M.__add(a, b)
  local r = Vector{}; for i = 1, #a do r[i] = a[i] + b[i] end
  return r
end

function _M.__sub(a, b)
  return __add(a, __unm(b))
end

function _M.__mul(a, b)
  if type(a) == 'number' then return b*a end
  if type(b) ~= 'number' then return nil end
  local r = Vector{}; for i = 1, #a do r[i] = a[i] * b end
  return r
end

function _M.__div(a, b)
  if type(b) ~= 'number' then return nil end
  local r = Vector{}; for i = 1, #a do r[i] = a[i] / b end
  return r
end

function _M.__unm(a, b)
  local r = Vector{}; for i = 1, #a do r[i] = -a[i] end
  return r
end

function _M.__eq(a, b)
  for i = 1, math.max(#a, #b) do if a[i] ~= b[i] then return false end end
  return true
end

function length(a)
  return sqrt(dot(a, a))
end

function cross(a, b)
  return Vector{ a.y*b.z - a.z*b.y, a.z*b.x - a.x*b.z, a.x*b.y - a.y*b.x }
end

function dot(a, b)
  local r = 0; for i = 1, #a do r = r + a[i]*b[i] end
  return r
end

function normalize(a)
  return a / a:length()
end
