
require 'Mask'
require 'Node'
require 'Event'
require 'Textures'
require 'Selection'

local _G = _G
local select = select
local SDL, gl, GL, glu = SDL, gl, GL, glu
local Mask, Node, Event, Textures, Selection =
  Mask, Node, Event, Textures, Selection

module(...)

-- TODO: all Windows should be inside a top-level object which:
-- * translates the coordinate system to the upper-left corner
-- * sets glStencilFunc(GL_EQUAL, 0, (default value))

local Window = _M
Node:subclass(Window)

Window.x = 0
Window.y = 0
Window.w = 0
Window.h = 0   -- negative number

Window.dragThreshold = 7

function Window:lostFocus()
  if self.pressed then return true end   -- keep focus
end

function Window:isDragged()
  return self.pressed and ((Event.mouse.x - self.pressed.x)^2 +
    (Event.mouse.y - self.pressed.y)^2) > self.dragThreshold^2
end

function Window:resize(w, h)
  self.w = w
  self.h = h
end

function Window:moveToFront()
  --[[
  if self.parent[#self.parent] ~= self then
    local parent, parentlen = self.parent, #self.parent
    self:replaceBy(nil); parent:addChild(self, parentlen) end
  ]]
  if self.parent[#self.parent] ~= self then self:refreshLayer() end
end

function Window:eventMouseMotion(ev)
  if self:isDragged() and self.pressed.resize then
    self:resize(self.pressed.origw - self.pressed.x + Event.mouse.x,
      self.pressed.origh + self.pressed.y - Event.mouse.y)
    return true
  end
end

function Window:eventMouseButtonAway(ev)
  if self.pressed then return self:eventMouseButton(ev) end
end

function Window:eventMouseButton(ev)
  if ev.down and ev.button == 1 then
    self:moveToFront()
    if self:getLoop().hover ~= self then return end
    self.pressed = { x = Event.mouse.x, y = Event.mouse.y,
      resize = ev.resize,   -- can be set in eventMouseButton by subclasses
      origw = self.w, origh = self.h }
    return true
  elseif not ev.down and ev.button == 1 then
    if self:isDragged() and not self.pressed.resize then
      self.x = self.x - self.pressed.x + Event.mouse.x
      self.y = self.y + self.pressed.y - Event.mouse.y
    end
    self.pressed = nil
    return true
  end
end

function Window:display()
  gl.PushMatrix()
  if self:isDragged() and not self.pressed.resize then
    gl.Translate(self.x - self.pressed.x + Event.mouse.cur_x,
      self.y + self.pressed.y - Event.mouse.cur_y, 0)
  else
    gl.Translate(self.x, self.y, 0)
  end
  Selection.push(self)
  Node.display(self)
  Selection.pop()
  gl.PopMatrix()
end

function Window:displayContent()
  gl.BindTexture(GL.TEXTURE_2D, 0)
  gl.Color(0, 1, 0)
  gl.Begin(GL.QUADS)
    gl.Vertex(0, 0)
    gl.Vertex(self.w, 0)
    gl.Vertex(self.w, self.h)
    gl.Vertex(0, self.h)
  gl.End()
  Node.displayContent(self)
end


Content = Node:subclass()

-- values relative to self.parent.x,y,w,h
Content.mx = 0   -- positive
Content.my = 0   -- negative
Content.mw = 0   -- negative (content is smaller than the whole window)
Content.mh = 0   -- positive

function Content:displayStencil()
  gl.Begin(GL.QUADS)
    gl.Vertex(0, 0)
    gl.Vertex(self.parent.w + self.mw, 0)
    gl.Vertex(self.parent.w + self.mw, self.parent.h + self.mh)
    gl.Vertex(0, self.parent.h + self.mh)
  gl.End()
end

function Content:displayContent()
  gl.BindTexture(GL.TEXTURE_2D, 0)
  gl.Color(0, 0, 1)
  if self:getLoop().focus == self then gl.Color(0, 0.5, 1) end
  gl.Begin(GL.TRIANGLES)
    gl.Vertex(-20, -20)
    gl.Vertex(self.parent.w / 2, self.parent.h)
    gl.Vertex(self.parent.w, self.parent.h / 2)
  gl.End()
  self:displayChildren()
end

function Content:display()
  gl.PushMatrix(); gl.Translate(self.mx, self.my, 0)
  Mask.Stencil.push(self)
  Selection.push(self)
  Node.display(self)
  Selection.pop()
  Mask.Stencil.pop()
  gl.PopMatrix()
end

function Content:eventMouseButton(ev)
  return true
end


TestRoot = Node.Root:subclass()

function TestRoot:display()
  Mask.Stencil.prepareTopLevel()
  Node.Root.display(self)
  Mask.Stencil.cleanupTopLevel()
end

