
Justify = {}

function Justify.splitWords(font, str)
  local words = {}
  for word in str:gmatch("%S+") do
    table.insert(words, { word = word, length = font:prepare(word) })
  end
  return words
end

function Justify.getLineEnd(font, words, line_width, line_begin)
  local space = font:prepare(" ")
  local total_length = 0
  if words[line_begin].length + space > line_width then return line_begin end
  for i = line_begin, #words do
    if total_length + words[i].length + space > line_width then return i-1 end
    total_length = total_length + words[i].length + space
  end
  return #words
end

function Justify.getSpaceSize(words, line_width, line_begin, line_end)
  for i = line_begin, line_end do
    line_width = line_width - words[i].length
  end
  -- num_spaces is num_words_on_line-1; num_words_on_line is line_end-line_begin+1
  return line_width / (line_end - line_begin)
end

function Justify.prepare(font, str, line_width, line_begin)
  line_begin = line_begin or 1
  local words = Justify.splitWords(font, str)
  local line_end = Justify.getLineEnd(font, words, line_width, line_begin)
  local prepare = {
    words = words, font = font, line_width = line_width,
    line_begin = line_begin, line_end = line_end,
    space = Justify.getSpaceSize(words, line_width, line_begin, line_end)
  }
  setmetatable(prepare, Justify)
  return prepare
end

function Justify.nextLine(prepare)
  if prepare.line_end == #prepare.words then return nil end
  local line_begin = prepare.line_end + 1
  local line_end = Justify.getLineEnd(prepare.font, prepare.words, prepare.line_width, line_begin)
  local nextLine = {
    words = prepare.words, font = prepare.font, line_width = prepare.line_width,
    line_begin = line_begin, line_end = line_end,
    space = Justify.getSpaceSize(prepare.words, prepare.line_width, line_begin, line_end)
  }
  setmetatable(nextLine, Justify)
  return nextLine
end

function Justify.prepareLines(font, str, line_width, line_begin)
  local line = Justify.prepare(font, str, line_width, line_begin)
  local lines = { line }
  while line do
    line = Justify.nextLine(line); table.insert(lines, line)
  end
  return lines
end

function Justify.print(prepare)
  for i = prepare.line_begin, prepare.line_end do
    prepare.font:print(prepare.words[i].word)
    if i ~= prepare.line_end then glTranslate(prepare.space, 0, 0) end
  end
end
