
require 'test.nehe.nehelib'

function setupWorld(world)
  local lines = {}
  for v in world:gmatch('[^\n]+') do
    if v:match('^ *[^/]') then lines[#lines+1] = v end
  end
  local numtriangles = tonumber(lines[1]:match('NUMPOLLIES (%d+)'))
  if not numtriangles then error'incorrect world file' end
  
  local lineno = 2
  local sector = {}
  for i = 1, numtriangles do
    sector[i] = {}
    for m = 1, 3 do
      local x,y,z,u,v = lines[lineno]:match('(%S+) +(%S+) +(%S+) +(%S+) +(%S+)')
      sector[i][m] = { x = tonumber(x), y = tonumber(y), z = tonumber(z), u = tonumber(u), v = tonumber(v) }
      lineno = lineno + 1
    end
  end
  return sector
end

heading, xpos, zpos, yrot, walkbias, walkbiasangle, lookupdown, z = 0, 0, 0, 0, 0, 0, 0, 0
blend = false
my_textures = { }
my_world = nil

function drawGLScene(interpolate)
  glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT)
  
  glLoadIdentity()
  glRotate(lookupdown, 1, 0, 0)
  glRotate(-heading, 0, 1, 0)
  glTranslate(-xpos, -walkbias-0.25, -zpos)
  
  glBindTexture(GL_TEXTURE_2D, my_textures[1])
  
  for tmp, t in ipairs(my_world) do
    glBegin(GL_TRIANGLES)
      glNormal(0, 0, 1)
      glTexCoord(t[1].u, t[1].v); glVertex(t[1].x, t[1].y, t[1].z)
      glTexCoord(t[2].u, t[2].v); glVertex(t[2].x, t[2].y, t[2].z)
      glTexCoord(t[3].u, t[3].v); glVertex(t[3].x, t[3].y, t[3].z)
    glEnd()
  end
end

function computeFrame()
  if Event.key[SDL.K_UP] then
    xpos = xpos - math.sin(math.rad(heading)) * 0.05
    zpos = zpos - math.cos(math.rad(heading)) * 0.05
    walkbiasangle = (walkbiasangle >= 359 and 0 or walkbiasangle+10)
    walkbias = math.sin(math.rad(walkbiasangle)) / 20
  end
  if Event.key[SDL.K_DOWN] then
    xpos = xpos + math.sin(math.rad(heading)) * 0.05
    zpos = zpos + math.cos(math.rad(heading)) * 0.05
    walkbiasangle = (walkbiasangle < 1 and 350 or walkbiasangle-10)
    walkbias = math.sin(math.rad(walkbiasangle)) / 20
  end
  if Event.key[SDL.K_RIGHT] then heading = heading - 1 end
  if Event.key[SDL.K_LEFT] then heading = heading + 1 end
  if Event.key[SDL.K_PAGEUP] then lookupdown = lookupdown - 1 end
  if Event.key[SDL.K_PAGEDOWN] then lookupdown = lookupdown + 1 end
end

initScreen()

glShadeModel(GL_SMOOTH)
glClearColor(0, 0, 0, 0)
glClearDepth(1)
glEnable(GL_DEPTH_TEST)
glDepthFunc(GL_LESS) -- TODO: why GL_LESS and not GL_LEQUAL?
glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST)
glEnable(GL_TEXTURE_2D)
glBlendFunc(GL_SRC_ALPHA, GL_ONE)

my_textures = { glGenTextures(1) }
glBindTexture(GL_TEXTURE_2D, my_textures[1])
glTexParameter(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR)
glTexParameter(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR)
loadImage('test/nehe/mud.bmp'):glTexImage2D()

my_world = setupWorld(assert(FS.load('test/nehe/world.txt'):read()))

NeheLoop:new():run()
