
-- based on <http://www.inf.puc-rio.br/~roberto/struct/teststruct>

local a,b,c,d,e,f,x

assert(#string.pack("C", 67324752) == 4)

assert(string.pack('a', 10) == "\010")
assert(string.pack('aaa', 10, 20, 30) == "\010\020\030")

assert(string.pack('<b', 10) == string.char(10, 0))
assert(string.pack('>b', 10) == string.char(0, 10))
assert(string.pack('<b', -10) == string.char(256-10, 256-1))

assert(string.pack('<c', 10) == string.char(10, 0, 0, 0))
assert(string.pack('>c', 10) == string.char(0, 0, 0, 10))
assert(string.pack('<c', -10) == string.char(256-10, 256-1, 256-1, 256-1))

assert(string.unpack(string.char(10, 0), '<b') == 10)
assert(string.unpack(string.char(0, 10), '>b') == 10)
assert(string.unpack(string.char(256-10, 256-1), '<b') == -10)

assert(string.unpack(string.char(10, 0, 0, 1), '<c') == 10 + 2^(3*8))
assert(string.unpack(string.char(0, 1, 0, 10), '>c') == 10 + 2^(2*8))
assert(string.unpack(string.char(256-10, 256-1, 256-1, 256-1), '<c') == -10)

-- limits
lims = {{'A', 255}, {'a', 127}, {'a', -128},
        {'B', 2^16 - 1}, {'b', 2^15 - 1}, {'b', -2^15},
        {'C', 2^32 - 1}, {'c', 2^31 - 1}, {'c', -2^31},
        -- 2^64-1 would work, but doesn't fit accurately in a lua_Number
        -- this is a rounding error, it can't be fixed by string.pack
        {'D', 2^63}, {'d', 2^62}, {'d', -2^62},
       }

if #string.pack('l', 1) == 4 then   -- 32-bit tests
  for _,v in pairs{
         {'H', 255}, {'h', 127}, {'h', -128},
         {'W', 2^16 - 1}, {'w', 2^15 - 1}, {'w', -2^15},
         {'I', 2^32 - 1}, {'i', 2^31 - 1}, {'i', -2^31},
         {'L', 2^32 - 1}, {'l', 2^31 - 1}, {'l', -2^31},
  } do table.insert(lims, v) end
end

for _, a in pairs{'', '>', '<'} do
  for _, l in pairs(lims) do
    local fmt = a .. l[1]
    assert(string.unpack(string.pack(fmt, l[2]), fmt) == l[2], fmt.." failed")
  end
end


-- tests for fixed-sized ints
for i, c in pairs{[1]='a',[2]='b',[4]='c'} do
  x = string.pack('<'..c, -3)
  assert(string.len(x) == i, i.." failed")
  assert(x == string.char(256-3) .. string.rep(string.char(256-1), i-1), i.." failed")
  assert(string.unpack(x, '<'..c) == -3, i.." failed")
end


-- alignment
d = string.pack("F", 5.1)
ali = {[1] = string.char(1)..d,
       [2] = string.char(1, 0)..d,
       [4] = string.char(1, 0, 0, 0)..d,
       [8] = string.char(1, 0, 0, 0, 0, 0, 0, 0)..d,
      }

for a,r in pairs(ali) do
  assert(string.pack(a.."!aF", 1, 5.1) == r, a.." failed")
  local x,y = string.unpack(r, a.."!aF")
  assert(x == 1 and y == 5.1, a.." failed")
end


-- strings
assert(string.unpack("alo alo", "s") == "")  -- unspecified length
assert(string.pack("s", "alo alo") == "alo alo")
assert(pcall(string.pack, "4s", "alo alo") == false)
assert(string.pack("z", "alo alo") == "alo alo\0")
assert(string.pack("4z", "alo alo") == "alo ")
assert(string.pack("5z", "alo alo") == "alo a")
assert(string.pack("10z", "alo alo") == "alo alo\0\0\0")
for _,fmt in pairs{'7s', 'z', '7z', '10z'} do
  assert(string.unpack(string.pack(fmt, "alo alo"), fmt) == "alo alo", fmt.." failed")
end
for _,fmt in pairs{'4z', '5z'} do   -- small lengths, incomplete information packed
  local x = string.pack(fmt, "alo alo")
  assert(string.unpack(x, fmt) == x, fmt.." failed")
end
assert(string.pack("4!a>7s", 1, "alo alo") == "\1alo alo")
assert(string.pack("2!<z", "alo alo") == "alo alo\0")
assert(string.pack("0s", "alo alo") == "alo alo")
for _, f in pairs{"H", "l", "b", "A"} do
  for _, s in pairs{"", "a", "alo", string.rep("x", 200)} do
    local x = string.pack("S"..f, s)
    assert(string.unpack(x, "S"..f) == s, f.." "..s:sub(1,5).." failed")
  end
end

-- tables
x = "\0\0\0\1\0\2\0\3alo\4\0"
a, b = string.unpack(x, "<x2bSab", true, 2)
assert(b == #x)
assert(table.concat(a, ',') == '1,2,alo,4' and a[3] == "alo")
assert(string.pack("<3x2bSab", a) == x)

-- offsets
x = string.pack("!>iiiii", 1, 2, 3, 4, 5)
local i, k = 0, 1
while i < #x do
  local v, j = string.unpack(x, "!>i", i)
  assert(j == i + 4 and v == k, k.." failed")
  i = j; k = k + 1
end

-- alignments are *not* relative to 'absolute' positions
x = string.pack("8! xd", 12)
assert(string.unpack(x, "8! 3xd") == 12)
assert(string.unpack(x, "8! d", 3) ~= 12)
x = string.pack("2! xb", 2)
assert(string.unpack(x, "2! xb") == 2)
assert(string.unpack(x, "2! b", 1) == 2*256)
assert(string.unpack(x, "2! b", 2) == 2)


assert(string.pack("<cbaxxB", -2, 10, -10, 250) ==
  "\254\255\255\255\010\000\246\000\000\250\000")

a,b,c,d = string.unpack("\254\255\255\255\010\000\246\000\000\250\000", "<cbaxxB")
assert(a == -2 and b == 10 and c == -10 and d == 250)

assert(string.pack(">cAxxB", -20, 10, 250) == "\255\255\255\236\10\0\0\0\250")

a, b, c, d = string.unpack("\255\255\255\236\10\0\0\0\250", ">cAxxB")
assert(a == -20 and b == 10 and c == 250 and d == 9)

a,b,c,d,e = string.unpack(
                  '000'..string.pack(">fFfB", 3.5, -24e-5, 200.5, 30000),
                  ">fFfB", 3)
assert(a == 3.5 and b == -24e-5 and c == 200.5 and d == 30000 and e == 21)

a,b,c,d,e = string.unpack(
                  '000'..string.pack("<fFxxfB", -13.5, 24e5, 200.5, 300),
                  "<fFxxfB", 3)
assert(a == -13.5 and b == 24e5 and c == 200.5 and d == 300 and e == 23)

x = string.pack(">BfcB", 10, 20, -30, 40001)
assert(string.len(x) == 2+4+4+2)
assert(string.unpack(x, ">f", 2) == 20)
a,b,c,d = string.unpack(x, ">bfcB")  -- "B"/"b" [sic]
assert(a == 10 and b == 20 and c == -30 and d == 40001)

local s = "hello hello"
x = string.pack("Sh",  s)
assert(string.unpack(x, "Sh") == s)
x = string.pack("SL", s)
assert(string.unpack(x, "  S L  ") == s)
x = string.pack("1z 3z a", s, s, 0)
assert(x == "hhel\0")
assert(string.unpack(x, "4xa") == 0)
assert(string.unpack(x, "z") == "hhel")
a, b = string.unpack(x.."\042", "za")
assert(a == "hhel" and b == 42)

if #string.pack('l', 1) == 4 and #string.pack('!xl', 1) == 8 then
  assert(string.pack("<!l", 3) == "\3\0\0\0")
  assert(string.pack("<!xl", 3) == "\0\0\0\0\3\0\0\0")
  assert(string.pack("<!xxl", 3) == "\0\0\0\0\3\0\0\0")
  assert(string.pack("<!xxxl", 3) == "\0\0\0\0\3\0\0\0")

  assert(string.unpack("\3\0\0\0", "<!l") == 3)
  assert(string.unpack("\0\0\0\0\3\0\0\0", "<!xl") == 3)
  assert(string.unpack("\0\0\0\0\3\0\0\0", "<!xxl") == 3)
  assert(string.unpack("\0\0\0\0\3\0\0\0", "<!xxxl") == 3)

  assert(string.pack("<2! h l w", 2, 3, 5) == "\2\0\3\0\0\0\5\0")
  a,b,c = string.unpack("\2\0\3\0\0\0\5\0", "<2!hlw")
  assert(a == 2 and b == 3 and c == 5)

  assert(string.pack("<8!hlw", 2, 3, 5) == "\2\0\0\0\3\0\0\0\5\0")
  a,b,c = string.unpack("\2\0\0\0\3\0\0\0\5\0", "<8!hlw")
  assert(a == 2 and b == 3 and c == 5)
else
  print "Skipping 32bit-only tests"
end

assert(string.pack(">zb", "aloi", 3) == "aloi\0\0\3")
assert(string.pack(">!zb", "aloi", 3) == "aloi\0\0\0\3")
x = "aloi\0\0\0\0\3\2\0\0"
a, b, c = string.unpack(x, "<!zc")
assert(a == "aloi" and b == 2*256+3 and c == string.len(x))

x = string.pack("4!zzz", "hi", "hello", "bye")
a,b,c = string.unpack(x, "zzz")
assert(a == "hi" and b == "hello" and c == "bye")
a, i = string.unpack(x, "z", 0)
assert(a == "hi")
a, i = string.unpack(x, "z", i)
assert(a == "hello")
a, i = string.unpack(x, "z", i)
assert(a == "bye")



-- test for weird conditions
assert(string.pack(">>>b <!!!<b", 10, 10) == "\000\010\010\000")
assert(string.pack("<3!xc", 10) == "\000\000\000\010\000\000\000")
assert(string.unpack("\2ab", "(comment)\nS\t(1a) A ()") == "ab")
assert(not pcall(string.pack, "Sf", "alo"))
assert(not pcall(string.pack, "S4L", "alo"))
assert(not pcall(string.pack, "3", 10))
assert(not pcall(string.pack, "aa", 6))
assert(not pcall(string.pack, "(comment "))
assert(not pcall(string.pack, ")"))
assert(select('#', string.unpack("\1\2\3\4\5", "aa\0aa")) == 3)
assert(pcall(string.pack, "", 1, 2, nil, 3, 4))
assert(not pcall(string.pack, "e", 10))
assert(string.pack("<=b", 8) == string.pack(">=b", 8))
assert(string.pack("") == "")
assert(string.pack("   ") == "")
assert(string.pack(">>><<<!!") == "")
assert(not pcall(string.unpack, "alo", "S"))
assert(not pcall(string.unpack, "alo", "z"))
assert(string.unpack("alo\0", "z") == "alo")
assert(not pcall(string.pack, "4s", "alo"))
assert(pcall(string.pack, "3s", "alo"))
assert(not pcall(string.unpack, "alo", "4s"))
assert(pcall(string.unpack, "alo", "3s"))
assert(not pcall(string.unpack, "\4alo", "Sh"))
assert(pcall(string.unpack, "\3alo", "Sh"))

assert(not pcall(string.unpack, "alo", "a", 4))
assert(not pcall(string.unpack, "alo", "a", 3))
assert(string.unpack("alo\3", "a", 3) == 3)

-- test lightuserdata access
if SDL then
  x = SDL.RWFromNewMem(8)
  string.pack(x:baseptr(), x:stopptr(), "<CSa", 192, "Abc")
  assert(assert(x:read(8)) == "\192\0\0\0\003Abc"); assert(x:seek('set'))
  a, b, c = string.unpack(x:baseptr(), x:stopptr(), "<CSa")
  assert(a == 192 and b == "Abc" and c == 8)
  a, b, c = string.unpack(x:baseptr(), 8, "<CSa")
  assert(a == 192 and b == "Abc" and c == 8)
  for i = 1, 7 do
    assert(x:seek('set', i))
    assert(not pcall(string.unpack, x:baseptr(), x:hereptr(), "<CSa"),
      i.." failed")
    assert(not pcall(string.unpack, x:baseptr(), i, "<CSa"),
      i.." failed")
    assert(not pcall(string.unpack, x:hereptr(), x:stopptr(), "<CSa"),
      i.." failed")
    assert(not pcall(string.unpack, x:baseptr(), x:stopptr(), "<CSa", i),
      i.." failed")
  end
else
  print "Skipping userdata tests"
end

print'OK'

