
#include <stdio.h>
#include <stdlib.h>

#include <lua.h>
#include <lauxlib.h>

#include <SDL_rwops.h>

#include "ftgl.h"


/* from lsdllib.c */
SDL_RWops *checkrwops (lua_State *L, int narg);
int rwops_disown_fn (lua_State *L);


static FT_Library get_ft_library(lua_State *L) {
  FT_Library *library_ud;
  lua_getfield(L, LUA_REGISTRYINDEX, "FTGL_FT_Library");
  library_ud = lua_touserdata(L, -1);
  lua_pop(L, 1);
  if(library_ud && *library_ud)
    return *library_ud;
  luaL_error(L, "can't find FTGL_FT_Library - FTGL didn't init correctly");
  return NULL;
}


/* Freetype reads from the font file even after FT_Open_Face finishes, so it has to be kept open. */
struct font_userdata_t {
  FTGL_Font *font;
  SDL_RWops *rw;
  FT_Open_Args args;
};

// SDL_RWops/FT_Stream interface is from SDL_ttf 2.0.9, available under GNU LGPL at www.libsdl.org
static unsigned long RWread(FT_Stream stream, unsigned long offset, unsigned char* buffer, unsigned long count) {
  SDL_RWops *src = (SDL_RWops *)stream->descriptor.pointer;
  SDL_RWseek(src, (int)offset, SEEK_SET);
  if(count == 0)
    return 0;
  return SDL_RWread(src, buffer, 1, (int)count);
}

static int ftgl_create(lua_State *L) {
  FT_Face face;
  struct font_userdata_t *handle;
  
  SDL_RWops *rw = checkrwops(L, 1);
  rwops_disown_fn(L);
  luaL_argcheck(L, SDL_RWtell(rw) >= 0, 1, "can't seek in RWops");
  luaL_checkinteger(L, 2); luaL_checkinteger(L, 3); luaL_checkinteger(L, 4);
  
  handle = (struct font_userdata_t *)lua_newuserdata(L, sizeof(struct font_userdata_t));
  handle->rw = rw;
  handle->args.flags = FT_OPEN_STREAM;
  handle->args.stream = (FT_Stream)malloc(sizeof(*handle->args.stream));
  if(handle->args.stream == NULL) luaL_error(L, "out of memory");
  memset(handle->args.stream, 0, sizeof(*handle->args.stream));
  handle->args.stream->read = RWread;
  handle->args.stream->descriptor.pointer = rw;
  SDL_RWseek(rw, 0, SEEK_END); handle->args.stream->size = SDL_RWtell(rw);
  SDL_RWseek(rw, 0, SEEK_SET);
  if(FT_Open_Face(get_ft_library(L), &handle->args, 0, &face) != 0)
    luaL_error(L, "error opening font face (FT_Open_Face failed)");
  if(FT_Set_Char_Size(face, 0, luaL_checkinteger(L, 2)*64, 0, 96) != 0)
    luaL_error(L, "error opening font face (FT_Set_Char_Size failed)");
  handle->font = FTGL_Create(face, luaL_checkinteger(L, 3), luaL_checkinteger(L, 4), lua_toboolean(L, 5));
  if(handle->font == NULL) luaL_error(L, "out of memory");
  luaL_getmetatable(L, "FTGL_Font"); lua_setmetatable(L, -2);
  return 1;
}


#if 0
static int my_ftgl_create(FT_Library ft, SDL_RWops *rw, struct font_userdata_t *handle, unsigned int font_size, int initial_texture_size, unsigned int texture_id, int use_kerning) {
  FT_Face face;
  
  handle->rw = rw;
  handle->args.flags = FT_OPEN_STREAM;
  handle->args.stream = (FT_Stream)malloc(sizeof(*handle->args.stream));
  if(handle->args.stream == NULL) return -1;
  memset(handle->args.stream, 0, sizeof(*handle->args.stream));
  handle->args.stream->read = RWread;
  handle->args.stream->descriptor.pointer = rw;
  SDL_RWseek(rw, 0, SEEK_END); handle->args.stream->size = SDL_RWtell(rw);
  SDL_RWseek(rw, 0, SEEK_SET);
  if(FT_Open_Face(ft, &handle->args, 0, &face) != 0)
    return -2;
  if(FT_Set_Char_Size(face, 0, font_size*64, 0, 96) != 0)
    return -3;
  handle->font = FTGL_Create(face, initial_texture_size, texture_id, use_kerning);
  if(handle->font == NULL) return -1;
  return 0;
}

static int ftgl_create(lua_State *L) {
  
  SDL_RWops *rw = checkrwops(L, 1);
  rwops_disown_fn(L);
  luaL_argcheck(L, SDL_RWtell(rw) >= 0, 1, "can't seek in RWops");
  
  struct font_userdata_t *handle = (struct font_userdata_t *)lua_newuserdata(L, sizeof(struct font_userdata_t));
  int status = my_ftgl_create(get_ft_library(L), rw, handle, luaL_checkinteger(L, 2), luaL_checkinteger(L, 3), luaL_checkinteger(L, 4));
  
  if(status == -1) luaL_error(L, "out of memory");
  else if(status == -2) luaL_error(L, "error opening font face (FT_Open_Face failed)");
  else if(status == -3) luaL_error(L, "error opening font face (FT_Set_Char_Size failed)");
  
  luaL_getmetatable(L, "FTGL_Font"); lua_setmetatable(L, -2);
  return 1;
}
#endif


static int ftgl_gc(lua_State *L) {
  struct font_userdata_t *handle = (struct font_userdata_t *)luaL_checkudata(L, 1, "FTGL_Font");
  if(handle->font) {
    if(handle->font->face)
      FT_Done_Face(handle->font->face);
    FTGL_Destroy(handle->font);
    handle->font = NULL;
  }
  if(handle->args.stream) {
    free((void *)handle->args.stream);
    handle->args.stream = NULL;
  }
  if(handle->rw) {
    SDL_RWclose(handle->rw);
    handle->rw = NULL;
  }
  return 0;
}


static FTGL_Font* getfont(lua_State *L) {
  struct font_userdata_t *handle = (struct font_userdata_t *)luaL_checkudata(L, 1, "FTGL_Font");
  luaL_argcheck(L, handle->font, 1, "expected FTGL font");
  return handle->font;
}


static int ftgl_index(lua_State *L) {
  const char *name = lua_tostring(L, 2);
  FTGL_Font *font = getfont(L);
  if(!lua_isstring(L, 2))
    return 0;
  luaL_getmetatable(L, "FTGL_Font");
  lua_pushvalue(L, 2); lua_rawget(L, -2);
  if(!lua_isnil(L, -1))
    return 1;
#define accessor(x,t) if(strcmp(name, #x) == 0) return lua_push##t (L, font->x), 1
  accessor(height, integer);
  accessor(ascent, integer);
  accessor(descent, integer);
  accessor(texture_size, integer);
  accessor(texture, integer);
  accessor(use_kerning, boolean);
  return 0;
}


static int ftgl_prepare(lua_State *L) {
  long result = FTGL_Prepare(getfont(L), luaL_checkstring(L, 2));
  if(result == -1) luaL_error(L, "out of memory");
  lua_pushinteger(L, result); return 1;
}


static int ftgl_print(lua_State *L) {
  int u = lua_isnumber(L, 3) ? lua_tointeger(L, 3) : (lua_isstring(L, 3) ? *lua_tostring(L, 3) : 0);
  FTGL_Print(getfont(L), luaL_checkstring(L, 2), u); return 0;
}


static int freetype_gc(lua_State *L) {
  FT_Library *library_ud = lua_touserdata(L, 1);
  if(*library_ud)
    FT_Done_FreeType(*library_ud);
  *library_ud = NULL;
  return 0;
}


int luaopen_ftgl(lua_State *L) {
  FT_Library *library_ud;
  const luaL_Reg reg_font[] = {
    { "prepare", ftgl_prepare },
    { "print", ftgl_print },
    { "__gc", ftgl_gc },
    { "__index", ftgl_index },
    { NULL, NULL }
  };

  lua_settop(L, 0);

  library_ud = (FT_Library *)lua_newuserdata(L, sizeof(FT_Library));
  if (FT_Init_FreeType(library_ud) != 0)
    luaL_error(L, "could not initialize Freetype");
  lua_newtable(L);
  lua_pushcfunction(L, freetype_gc);
  lua_setfield(L, -2, "__gc");
  lua_setmetatable(L, -2);
  lua_setfield(L, LUA_REGISTRYINDEX, "FTGL_FT_Library");

  luaL_getmetatable(L, "SDL_RWops");
  if(!lua_istable(L, 1))
    luaL_error(L, "SDL not found (required by FTGL)");
  lua_pushcfunction(L, ftgl_create);
  lua_setfield(L, 1, "font");
  
  luaL_newmetatable(L, "FTGL_Font");
  luaL_register(L, NULL, reg_font);

  lua_settop(L, 1);
  return 1;
}

